; defining registers of PIC16F8x
; system wide registers and vars
(define W 0)
(define stack '())
(define mem (make-vector 80 0)) ; all registers, file registers starts at 0x0C
(define skip 0)  ; used for tests
;(define WDT 0)   ; Watchdog Timer
(define NO-OP 0)   ; used if an instruction requires 2 cycles
(define sleeping #t) ; idle - wait for interrupt or hardware reset
(define WDTE 0)  ; Configuration bit (enable or not the WDT) normally 1

(vector-set! mem 3 (string->number "0011000" 2))   ; OPTION


; BANK 0  (00h)
(define (INDF) (vector-ref mem 0))
(define (TMR0) (vector-ref mem 1))
(define (PCL) (vector-ref mem 2))
(define (STATUS) (vector-ref mem 3))
(define FSR INDF)   ; pas test encore
(define (PORTA) (vector-ref mem 5))  ; 5 bits
(define (PORTB) (vector-ref mem 6))
(define (EEDATA) (vector-ref mem 8))
(define (EEADR) (vector-ref mem 9) )
(define (PCLATH) (vector-ref mem 10)) ; 5 bits
(define (INTCON) (vector-ref mem 11))

; BANK 1  (80h)
(define OPTION_REG 255)
(define TRISA 31) ; 5 bits
(define TRISB 255)


; Operations on Program Counter

(define PC-ref (lambda () (+ (PCL) (* 256 (PCLATH)))))    ; returns value of in registers

(define PC-set!     ; sets a new PC, in PCL and PCLATH
  (lambda (pc)
    (begin (vector-set! mem 2 (bitwise-and 255 pc))
           (vector-set! mem 10 (+ (bitwise-and (PCLATH) 24)  ; PCLATH<4:3>
                                  (bitwise-and 7 (/ (bitwise-and 1792 pc) 256))))))) ; PCLATH<2:0>


; ------ STACK ------

(define TOS   ; pops and returns
  (lambda ()
    (if (null? STACK)
        (error "void stack")
        (let ((rval (car STACK)))
          (begin (set! STACK (cdr STACK))
                 rval)))))

(define push
  (lambda (o)
    (set! stack (cons o stack))))


; file registers I/O access
; takes address as a hexadecimal unprefixed string (like 2c)

(define file-set!    ; sets file register adress to data
  (lambda (address data)
    (let ((decaddr (if (string? address) (string->number address 16) address)))
      (cond ((< decaddr 80) (vector-set! mem decaddr data))
            ((= decaddr 129) (set! OPTION_REG data))
            ((= decaddr 133) (set! TRISA data))
            ((= decaddr 134) (set! TRISB data))
            (else #f)))))

(define file-ref     ; get value in file register adress
  (lambda (address)
    (let ((decaddr (if (string? address) (string->number address 16) address)))
      (cond ((< decaddr 80) (vector-ref mem decaddr))
            ((= decaddr 129) OPTION_REG)
            ((= decaddr 133) TRISA)
            ((= decaddr 134) TRISB)
            (else #f)))))


; bit operations

(define is-bit-set?
  (lambda (b L)
    (let ((pow2 (expt 2 b)))
      (= pow2 (bitwise-and L pow2)))))

(define bit-set
  (lambda (b L)
    (bitwise-ior L (expt 2 b))))

(define bit-set!
  (lambda (b f)
    (file-set! f (bit-set b (file-ref f)))))

(define bit-clear
  (lambda (b L)
    (bitwise-and L (- 255 (expt 2 b)))))

(define bit-clear!
  (lambda (b f)
    (file-set! f (bit-clear b (file-ref f)))))

(define Z-set!
  (lambda ()
    (vector-set! mem 3 (bit-set 2 (STATUS)))))

(define Z-clear!
  (lambda ()
    (vector-set! mem 3 (bit-clear 2 (STATUS)))))

(define C-set!
  (lambda ()
    (vector-set! mem 3 (bit-set 0 (STATUS)))))

(define C-clear!
  (lambda ()
    (vector-set! mem 3 (bit-clear 0 (STATUS)))))


;------------- STRINGS OPERATIONS ---------------
; used for display, adds obsolete 0
; takes an integer and returns the 2 chars string
(define hexa-string
  (lambda (n)
    (if (< n 16)
        (string-append "0" (number->string n 16))
        (number->string n 16))))

; returns 8 chars string
(define binary-string
  (lambda (n len)
    (let ((original-string (number->string n 2)))
      (letrec ((iter (lambda (i res)
                       (if (>= i len)
                           res
                           (iter (+ i 1) (string-append "0" res))))))
        (iter (string-length original-string) original-string)))))


; remove 0x if passed string begins with this,
; returns decimal representation of the number in hex.
(define remove-0x
  (lambda (str)
    (let ((do-it (string->number str 16)))
      (if do-it
          do-it
          (string->number (substring str 2 (string-length str)) 16)))))


;-------------------- DEBUG ---------------------
(define Z
  (lambda ()
    (bitwise-and (STATUS) 4)))

(define C
  (lambda ()
    (bitwise-and (STATUS) 1)))